#define _CRT_SECURE_NO_WARNINGS
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>
#include <stdlib.h>
#include <tuple>
#include "heapStruct.h"
#include "structDef.h"
#include "dataRead.h"
#include "eppsteinStruct.h"
#include "heuristiqueDep.h"
#include <algorithm>
#include <cmath>
#include <set>
#include <random>
using namespace std;

vector<int> generateNumbers(int n) {
    vector<int> nums(n);
    for (int i = 1; i <= n; i++) {
        nums[i - 1] = i;
    }
    return nums;
}

vector<vector<int>> generateRandomPermutationsSet() {
    vector<vector<int>> permutationsSet;
    vector<int> nums = generateNumbers(31);

    // Use a uniform distribution to generate random numbers
    std::random_device rd;
    std::mt19937 g(rd());

    for (int i = 0; i < 1; i++) {
        // Shuffle the number vector with a random permutation function
        std::shuffle(nums.begin(), nums.end(), g);

        // Generate a random permutation from the shuffled vector
        permutationsSet.push_back(nums); // Only need the first generated permutation
    }

    return permutationsSet;
}

int main() {
    // Parallel deployment parameter (true for parallel, false for sequential)
    int para = false;
    // Balanced deployment parameter (true for balanced, false for unbalanced)
    int depEqui = true;
    // Number of iterations for learning
    int nbrIttApp = 0;
    // Parallelism after deployment (if on the same node)
    int paraAfter = 0;

    // Internal parallelism
    bool paraInt = true;

    // Jain's fairness index
    double jainInd = -1;

    // No current version to deploy SFCs in parallel while optimizing balancing
    if (para == true && depEqui == true) {
        cout << endl << "No current version to deploy SFCs in parallel while optimizing balancing" << endl;
        return(0);
    }

    // Loop to read from file P and go through folder instance I with I going through the following loop
    string p = "E";

    for (int ii = 1; ii < 10; ii++) {
        cout << endl << endl << endl << "Scenario: " << ii << endl << endl << endl;

        // Data structures for graphs
        vector<node> listNode;
        vector<node> listNodeOrig;
        vector<node> listNodeSauv;
        vector<node> eppsteinGraph;
        vector<int> path;
        vector<deployment> resDep;
        vector<deployment> resDepI;
        // Data structures for service requests
        vector<int> F;
        vector<vector<tuple<int, vector<int>>>> Q; // set of service requests
        vector<int> latenceDemande; // maximum latency per service request
        vector<int> sources; // source related to each service request
        vector<int> destinations; // destination related to each service request
        vector<int> cl; // class of the SFC
        int nbBR; // number of BE SFCs
        int nbLLT; // number of LLT SFCs
        vector<vector<int>> kPaths;
        vector<vector<vector<int>>> paraGen;

        // Metrics and indicators to calculate
        int sumLatSFC = 0;
        double deploymentSimpleDuration;
        double parallelisationDuration3;
        double deploymentDuration1;
        int nmbrDeployedSFC4;
        double deploymentMediumSfc2;
        int nmbrSFCrespLatence;
        int nmbrSFCdepassLat5;
        int moyLatSFC6;
        int ecartType7;

        // Loop to determine the number of instances to read (repeat) within the instance and within the SFC file
        for (int in = 0; in < 1; in++) {
            cout << endl << endl << endl << "Instance: " << in << endl << endl << endl;

            // I/O streams
            ofstream fluxRes;
            ofstream fluxRes2;

            // Reading data
            clock_t begin = clock();
            // Reading the initial graph for the shortest path
			string pathT = "C:/Users/magnouch/Documents/Heuristique/Instance";
			string const pathS = pathT + p + "/Instance" + to_string(ii); 
			listNode = graphReader(pathS);
			listNodeOrig = listNode;
			listNodeSauv = listNode;

			// Reading the SFCs (Service Function Chains)
			dataReader(pathS, &F, &Q, &latenceDemande, &sources, &destinations, &cl, &nbBR, &nbLLT, &paraGen, &fluxRes, &fluxRes2, in, ii); 

			// Creating a vector to hold SFCs with an associated score
			vector<scoreSFC> orderedSFC;
			vector<scoreSFC> orderedSFCI;

			// Assigning a score based on the margin
			orderedSFC = attribuerScoreMarge(Q, latenceDemande, sources, destinations);

			clock_t begin; 
			// orderedSFC = attribuerScoreApp(Q, nbrIttApp,latenceDemande, sources, destinations);

			clock_t start = clock();

			// Ordering the SFCs 
			sort(orderedSFC.begin(), orderedSFC.end(), minMax);

			// Deploying the SFCs
			if (para) {
				resDep = deployementSPara(orderedSFC, &listNode, in); 
			} else {
				resDep = deployementS(orderedSFC, &listNode, in, depEqui, paraInt);
			}

			listNodeSauv = listNode;

			// Saving the number of SFCs that exceed latency limits
			nmbrDeployedSFC_FCT_APP = deployedSFC(resDep);	
			nmbrDeployedSFC_FCT_APP_Last = deployedSFC(resDep);
			nmbrSFCrespLatence_FCT_APP = SFCresLat(listNode, resDep, orderedSFC, &sumLatSFC_FCT_APP);
			nmbrSFCrespLatence_FCT_APP_Last = nmbrSFCrespLatence_FCT_APP; 
			nmbrSFCdepassLatLast = nmbrDeployedSFC_FCT_APP - nmbrSFCrespLatence_FCT_APP;
			orderedSFCI = orderedSFC;
			jainInd_Last = jain_Ind_cal(&listNode);

			// Learning loop to adjust the order of processing SFCs based on latency exceedance
			int idS = -1;
			int scoreS = -1;
			for (int t = 0; t < nbrIttApp; t++) {
				sort(orderedSFC.begin(), orderedSFC.end(), minMax);

				if ((orderedSFC[0].id == idS) && (orderedSFC[0].score < scoreS)) {
					orderedSFC[0].score += nbrIttApp;
				}

				sort(orderedSFC.begin(), orderedSFC.end(), minMax);

				idS = orderedSFC[0].id;
				scoreS = orderedSFC[0].score;
				listNode = listNodeOrig;

				// Deploying SFCs
				if (para) {
					resDepI = deployementSPara(orderedSFC, &listNode, in);
				} else {
					resDepI = deployementS(orderedSFC, &listNode, in, depEqui, paraInt);
				}

				// Adjusting the score for each SFC based on whether latency requirements are met
				for (int q = 0; q < resDepI.size(); q++) {
					if (latSFC(listNode, resDepI[q].chain, resDepI[q].placement) > latenceDemande[q]) {
						orderedSFC[q].score--;
					} else if (latSFC(listNode, resDepI[q].chain, resDepI[q].placement) < latenceDemande[q]) {
						orderedSFC[q].score++;
					}
				}

				// Storing the scenario with the best performance in terms of deployed SFCs and adherence to latency limits
				nmbrDeployedSFC_FCT_APP = deployedSFC(resDepI);
				nmbrSFCrespLatence_FCT_APP = SFCresLat(listNode, resDepI, orderedSFC, &sumLatSFC_FCT_APP);
				nmbrSFCdepassLat_FCT_APP = nmbrDeployedSFC_FCT_APP - nmbrSFCrespLatence_FCT_APP;
				jainInd = jain_Ind_cal(&listNode);

				// Selecting scenarios with the most latency-compliant SFCs
				if (nmbrSFCrespLatence_FCT_APP > nmbrSFCrespLatence_FCT_APP_Last) {
					resDep = resDepI;
					orderedSFCI = orderedSFC;
					listNodeSauv = listNode;
					nmbrDeployedSFC_FCT_APP_Last = nmbrDeployedSFC_FCT_APP;
					nmbrSFCrespLatence_FCT_APP_Last = nmbrSFCrespLatence_FCT_APP;
					nmbrSFCdepassLatLast = nmbrSFCdepassLat_FCT_APP;
					jainInd_Last = jainInd;
				}
				else if (nmbrSFCrespLatence_FCT_APP == nmbrSFCrespLatence_FCT_APP_Last) {
					if (nmbrDeployedSFC_FCT_APP > nmbrDeployedSFC_FCT_APP_Last) {
						resDep = resDepI;
						nmbrSFCdepassLatLast = nmbrSFCdepassLat_FCT_APP;
						orderedSFCI = orderedSFC;
						listNodeSauv = listNode;
						nmbrDeployedSFC_FCT_APP_Last = nmbrDeployedSFC_FCT_APP;
						nmbrSFCrespLatence_FCT_APP_Last = nmbrSFCrespLatence_FCT_APP;
						nmbrSFCdepassLatLast = nmbrSFCdepassLat_FCT_APP;
						jainInd_Last = jainInd;
					}
					else if (nmbrDeployedSFC_FCT_APP > nmbrDeployedSFC_FCT_APP_Last) {
						if (jainInd > jainInd_Last)
							resDep = resDepI;
							nmbrSFCdepassLatLast = nmbrSFCdepassLat_FCT_APP;
							orderedSFCI = orderedSFC;
							listNodeSauv = listNode;
							nmbrDeployedSFC_FCT_APP_Last = nmbrDeployedSFC_FCT_APP;
							nmbrSFCrespLatence_FCT_APP_Last = nmbrSFCrespLatence_FCT_APP;
							nmbrSFCdepassLatLast = nmbrSFCdepassLat_FCT_APP;
							jainInd_Last = jainInd; 
					}
				}
			}
			
			clock_t end = clock();
			clock_t start2 = clock();

			// Parallelization in case an SFC does not meet latency requirements
			if (paraAfter) {
				for (int q = 0; q < orderedSFCI.size(); q++) {
					if (latSFC(listNode, resDep[q].chain, resDep[q].placement) > orderedSFCI[q].lat) {
						paraInterF(&resDep[q], listNode, orderedSFCI[q].SFC);
					}
				}
			}

			clock_t end2 = clock();

			// Calculate various timing metrics
			sumLatSFC = 0;
			deploymentSimpleDuration = double(end - start) / CLOCKS_PER_SEC;
			parallelisationDuration3 = double(end2 - start2) / CLOCKS_PER_SEC;
			deploymentDuration1 = deploymentSimpleDuration + parallelisationDuration3;
			nmbrDeployedSFC4 = deployedSFC(resDep);
			deploymentMediumSfc2 = deploymentDuration1 / nmbrDeployedSFC4;
			nmbrSFCrespLatence = SFCresLat(listNodeSauv, resDep, orderedSFCI, &sumLatSFC);
			nmbrSFCdepassLat5 = nmbrDeployedSFC4 - nmbrSFCrespLatence;

			// Display deployment results
			cout << "\n-------Deployed: " << nmbrDeployedSFC4 << endl;
			cout << "\n-------Resp Lat : " << nmbrSFCrespLatence << endl;

			// Calculate average latency per SFC
			if (nmbrDeployedSFC4 != 0) {
				moyLatSFC6 = sumLatSFC / nmbrDeployedSFC4;
			} else {
				moyLatSFC6 = -1;
			}
			ecartType7 = sqrt(sommeLatMinusMoy(listNodeSauv, resDep, moyLatSFC6) / (nmbrDeployedSFC4 - 1));

			// Define current time
			time_t current_time = time(nullptr);
			tm* local_time = localtime(&current_time);

			// Write scenario information to file
			fluxRes2 << "\nScnario: " << ii << "; Instance: " << in + 1 << ";" << p << "; Time: "
					 << local_time->tm_hour << "h" << local_time->tm_min << "m" << local_time->tm_sec
					 << "s ; " << jainInd_Last << endl; // Added instance index

			// Print detailed results for SFC deployment
			for (int q = 0; q < resDep.size(); q++) {
				for (int q = 0; q < resDep.size(); q++) {
							cout << endl << "Dtaills de la solution pour le dploiement de la SFC : " << q << ", ID : " << orderedSFCI[q].id << "; avec latence arc : " << latArcF(listNodeSauv, resDep[q].chain) << "; Et la latence de SFC : " << latEx(resDep[q].placement) << "; Et la latence total : " << latSFC(listNodeSauv, resDep[q].chain, resDep[q].placement) << "; Latence requise : " << orderedSFCI[q].lat << endl;
							fluxRes2 << endl << "SFC : " << orderedSFCI[q].id << ", ID: " << orderedSFCI[q].id << "; Arc Latency: " << latArcF(listNodeSauv, resDep[q].chain) << "; Execution Latency: " << latEx(resDep[q].placement) << "; Total Latency: " << latSFC(listNodeSauv, resDep[q].chain, resDep[q].placement) << "; Latence requise : " << orderedSFCI[q].lat << endl;
							for (int i = 0; i < resDep[q].placement.size(); i++) {
								cout << endl << "Sur le noeud : " << get<0>(resDep[q].placement[i]) << " sont deploye les micro-services : " << endl;
								fluxRes2 << "Node: " << get<0>(resDep[q].placement[i]) << "; Are deployed ==> ";
								for (int j = 0; j < get<1>(resDep[q].placement[i]).size(); j++) {
									cout << get<1>(resDep[q].placement[i])[j] << ", ";
									fluxRes2 << get<1>(resDep[q].placement[i])[j] << ", ";
								}
								//Affichage des ensembles de paralllism de la requete 
								cout << endl << "Affichage des ensembles de parallelisme " << endl;
								cout << endl << "Ens : " << i << " , est : ";
								for (int j = 0; j < get<2>(resDep[q].placement[i]).size(); j++) {
									fluxRes2 << endl << "Parallelisation set: " << j << ", Contains: ";
									for (int y = 0; y < get<2>(resDep[q].placement[i])[j].size(); y++) {
										cout << get<2>(resDep[q].placement[i])[j][y] << ", ";
										fluxRes2 << get<2>(resDep[q].placement[i])[j][y] << ", ";
									}
								}

							}

						}
			}

			// Display final parameters to be output to a file
			cout << "\n\nTotal deployment time: " << deploymentDuration1 << endl;
			cout << "\nAverage deployment time: " << deploymentMediumSfc2 << endl;
			cout << "\nParallelization time: " << parallelisationDuration3 << endl;
			cout << "\nNumber of deployed SFCs: " << nmbrDeployedSFC4 << endl;
			cout << "\nNumber of SFCs exceeding latency: " << nmbrSFCdepassLat5 << endl;
			cout << "\nAverage latency per SFC: " << moyLatSFC6 << endl;
			cout << "\nStandard deviation of SFC latency: " << ecartType7 << endl;

			// Write the final parameters to the output file
			fluxRes << "\n" << p << " ; " << ii << " ; " << in + 1 << " ; "
					<< local_time->tm_hour << "h" << local_time->tm_min << "m" << local_time->tm_sec << "s"
					<< " ; " << deploymentDuration1 << " ; " << deploymentMediumSfc2 << " ; "
					<< parallelisationDuration3 << " ; " << nmbrDeployedSFC4 << " ; "
					<< nmbrSFCdepassLat5 << " ; " << moyLatSFC6 << " ; " << ecartType7 << ";"
					<< jainInd_Last;

		}
	}

	return 0;
}


