#include<iostream>
#include<climits>
#include<tuple>
#include<string>
#include<vector>
#include <fstream>;
#include <sstream>;
using namespace std; 
#pragma once

typedef tuple<int, int, vector<int>> tup;

// Preprocesses SFCs to determine which nodes can be parallelized
vector<vector<int>> calculateParallelism(vector<tuple<int, vector<int>>> SFC, int nbMS) {
    vector<vector<int>> parallelismMatrix;
    bool isParallel;
    for (int i = 0; i < nbMS; i++) {
        vector<int> intermediate;
        for (int j = 0; j < nbMS; j++) {
            isParallel = true;
            // Check if i and j are not part of the same SFC
            for (int f = 0; f < SFC.size(); f++) {
                if (get<0>(SFC[f]) == i) {
                    for (int f2 = 0; f2 < get<1>(SFC[f]).size(); f2++) {
                        if (j == get<1>(SFC[f])[f2]) {
                            isParallel = false;
                        }
                    }
                }
            }
            // Check if j and i are not part of the same SFC
            for (int f = 0; f < SFC.size(); f++) {
                if (get<0>(SFC[f]) == j) {
                    for (int f2 = 0; f2 < get<1>(SFC[f]).size(); f2++) {
                        if (i == get<1>(SFC[f])[f2]) {
                            isParallel = false;
                        }
                    }
                }
            }

            // Add 1 for parallelizable, 0 for not
            intermediate.push_back(isParallel ? 1 : 0);
        }
        parallelismMatrix.push_back(intermediate);
    }
    return parallelismMatrix;
}

// Function to split a string by a delimiter character
vector<string> split(const string& s, char delim) {
    stringstream ss(s);
    string item;
    vector<string> elements;
    while (getline(ss, item, delim)) {
        elements.push_back(item);
    }
    return elements;
}

// Reads the graph from files and creates a list of nodes
vector<node> graphReader(string path) {
    vector<node> listNodeGR;
    string const nodeFilePath = path + "/instanceNode.txt";
    string const arcFilePath = path + "/instanceARC.txt";
    ifstream nodeFile(nodeFilePath.c_str());
    ifstream arcFile(arcFilePath.c_str());

    // Reading nodes
    string line;
    while (getline(nodeFile, line)) {
        vector<string> x = split(line, ',');
        node nodeGR;
        nodeGR.id = stoi(x[0]);
        nodeGR.capaMem = stoi(x[1]);
        listNodeGR.push_back(nodeGR);
    }

    // Reading arcs and updating the successors of each node
    while (getline(arcFile, line)) {
        node* nodeA = NULL; // Pointer to the source node of the arc
        node* nodeB = NULL; // Pointer to the destination node of the arc
        int weight = -1;
        vector<string> x = split(line, ',');
        weight = stoi(x[2]);
        for (int i = 0; i < listNodeGR.size(); i++) {
            if (listNodeGR[i].id == stoi(x[0]))
                nodeA = &listNodeGR[i];
            if (listNodeGR[i].id == stoi(x[1]))
                nodeB = &listNodeGR[i];
        }
        if ((nodeA == NULL) || (nodeB == NULL))
            cout << endl << "ERROR: Node not found" << endl;
        tuple<node*, int> arc(nodeB, weight);

        nodeA->succ.push_back(arc);
    }
    return listNodeGR;
}


// This function reads data from files and initializes vectors and variables accordingly
void dataReader(
    string path,
    vector<int>* F,
    vector<vector<tuple<int, vector<int>>>>* Q,
    vector<int>* latenceDemande,
    vector<int>* source,
    vector<int>* destination,
    vector<int>* cl,
    int* nbBR,
    int* nbLLT,
    vector<vector<vector<int>>>* paraGen,
    ofstream* fluxRes,
    ofstream* fluxRes2,
    int ni,
    int nbSc
) {
    // Define the paths to the instance files
    string const micro = path + "/instanceMicroService.txt";
    string const SFC = path + "/instanceSFC.txt";
    string ress("C:/Users/magnouch/Documents/Heuristique");
    string const res = ress + "/Resultats/resultatSummary.txt";
    string const res2 = ress + "/Resultats/resultatDetails.txt";

    // Open the instance files
    ifstream fluxSFC(SFC.c_str());
    ifstream fluxMicroService(micro.c_str());
    ifstream fluxTabPara("C:/Users/magnouch/Documents/Modlisation/Autre/instance/tabPara.txt");
    ifstream fluxTabEqui("C:/Users/magnouch/Documents/Modlisation/Autre/instance/tabRemp.txt");
    fluxRes->open(res.c_str(), ios::app);
    fluxRes2->open(res2.c_str(), ios::app);

    // Read data from the microservice file
    string ligne;
    while (getline(fluxMicroService, ligne)) {
        F->push_back(stoi(ligne));
    }

    // Skip lines in the SFC file based on 'ni'
    for (int t = 0; t < ni * 5; t++) {
        getline(fluxSFC, ligne); 
    }

    // Determine the number of SFCs to read from the file based on 'nbSc'
    int nbT = determineNbT(nbSc);  // This should be a function to determine the number of SFCs based on 'nbSc'

    // Read and parse SFC data
    for (int t = 0; t < nbT; t++) {
        getline(fluxSFC, ligne);
        if (ligne[0] != 'x') {
            vector<string> SFC2 = split(ligne, ',');
            vector<tuple<int, vector<int>>> inter3;
            for (int j = 0; j < SFC2.size() - 1; j++) {
                vector<string> micro = split(SFC2[j], '\\');
                int i = stoi(micro[0]);
                vector<int> inter2;
                for (int k = 1; k < micro.size(); k++) {
                    inter2.push_back(stoi(micro[k]));
                }
                inter3.push_back(make_tuple(i, inter2));
            }
            Q->push_back(inter3);
            vector<string> x2 = split(SFC2.back(), ';');
            source->push_back(stoi(x2[0]));
            destination->push_back(stoi(x2[1]));
            latenceDemande->push_back(stoi(x2[2]));
            cl->push_back(stoi(x2[4]));
        }
    }

    // Read and parse parallelism table data
    vector<vector<int>> tabPara;
    while (getline(fluxTabPara, ligne)) {
        vector<string> x = split(ligne, ',');
        vector<int> inter;
        for (const string& val : x) {
            inter.push_back(stoi(val));
        }
        tabPara.push_back(inter);
    }

    // Read and parse equivalence table data
    vector<tuple<int, int>> equiv;
    while (getline(fluxTabEqui, ligne)) {
        vector<string> x = split(ligne, ',');
        equiv.push_back(make_tuple(stoi(x[2]), stoi(x[0])));
    }

    // Initialize 'paraGen' which represents the parallelism between two service functions within an SFC
    for (const auto& q : *Q) {
        paraGen->push_back(calculPara(q, F->size()));
    }

    // Initialize the variable 'nbBR' which represents the number of BE SFCs
    *nbBR = 0;
    for (int q = 0; q < Q->size(); q++) {
        if ((*cl)[q] == 0) {
            (*nbBR)++;
        }
    }

    // Initialize the variable 'nbLLT' which represents the number of LLT SFCs
    *nbLLT = 0;
    for (int q = 0; q < Q->size(); q++) {
        if ((*cl)[q] == 1) {
            (*nbLLT)++;
        }
    }
}

// Function to print node capacities
void printNodeCapa(vector<node>* listNode) {
    int sum = 0; 
    for (int no = 0; no < listNode->size(); no++) {
        cout << "Node : " << (*listNode)[no].id << ", Capa = " << (*listNode)[no].capaMem << endl; 
        sum += (*listNode)[no].capaMem; 
    }
    cout << endl << "Somme : " << sum << endl; 
}

// Function to calculate Jain's fairness index
double jain_index(const vector<double>& resources) {
    double sum = accumulate(resources.begin(), resources.end(), 0.0);
    double sum_squared = inner_product(resources.begin(), resources.end(), resources.begin(), 0.0);
    int n = resources.size();
    return (sum * sum) / (n * sum_squared);
}

// Function to calculate and display Jain's index
double jain_Ind_cal(vector<node>* listNode) {
    vector<double> deploymentVec;
    for (int no = 0; no < listNode->size(); no++) {
        deploymentVec.push_back(static_cast<double>((*listNode)[no].capaMem));
    }
    return jain_index(deploymentVec);
}

int determineNbT(int nbSc) {
    switch (nbSc) {
        case 1: return 25;
        case 2: return 50;
        case 3: return 100;
        case 4: return 250;
        case 5: return 500;
        case 6: return 1000;
        case 7: return 2500;
        case 8: return 5000;
        case 9: return 10000;
        default: return 0; // or some default number if nbSc is not in [1,9]
    }
}
