#pragma once
#include<iostream>
#include<climits>
#include<tuple>
#include<string>
#include<vector>

using tuple_t = std::tuple <node*, int>;

// Function to erase a tuple_t from a list of tuple_t
void eraseSucc(vector<tuple_t>* listSucc, tuple_t succ) {
    vector<tuple_t>::iterator it = (*listSucc).begin();

    // Iterate until the element to erase is found
    while (*it != succ) {
        it++;
    }

    // Erase the element from the list
    (*listSucc).erase(it);
}

// Function to erase a vector of ints from a list of vectors
void eraseVec(vector<vector<int>>* list, vector<int> val) {
    vector<vector<int>>::iterator it = (*list).begin();

    // Iterate until the element to erase is found
    while (*it != val) {
        it++;
    }

    // Erase the element from the list
    (*list).erase(it);
}

// Function to erase an arc from a list of arcs
void eraseVecNodeS(vector<arc>* list, int val) {
    vector<arc>::iterator it = (*list).begin();

    // Iterate until the arc with the specified id is found
    while ((*it).id != val) {
        it++;
    }

    // Erase the arc from the list
    (*list).erase(it);
}

// Function to erase a treeNode from a list of treeNodes
void eraseTreeNodeVect(vector<treeNode*>* list, treeNode val) {
    vector<treeNode*>::iterator it = (*list).begin();

    // Iterate until the treeNode with the specified id is found
    while ((*it)->id != val.id) {
        it++;
    }

    // Erase the treeNode from the list
    (*list).erase(it);
}

// Function to invert a graph: reverse the direction of arcs
vector<node> graphInverse(vector<node> listNode) {
    vector<node> listNodeT; // Temporary list to store the inverted nodes

    // Copy the nodes with capacities but without successors
    for (int i = 0; i < listNode.size(); i++) {
        node n;
        n.capaMem = listNode[i].capaMem;
        n.id = listNode[i].id;
        n.dis = listNode[i].dis;
        n.preNex = listNode[i].preNex;
        listNodeT.push_back(n);
    }
    
    // NOTE: NODE IDs MUST CORRESPOND TO THEIR ORDER IN THE NODE FILE; EX: NODE 3 AT POSITION 3
    // Reverse the arcs
    for (int i = 0; i < listNode.size(); i++) {
        for (int j = 0; j < listNode[i].succ.size(); j++) {
            tuple <node*, int> t(&listNodeT[i], get<1>(listNode[i].succ[j]));
            listNodeT[get<0>(listNode[i].succ[j])->id].succ.push_back(t);
        }
    }

    return listNodeT;
}

// Function to invert a graph: reverse the direction of arcs and return a pointer to the new graph
vector<node>* graphInverse2(const vector<node>& listNode) {
    vector<node>* listNodeT = new vector<node>(listNode.size());

    // Copy the nodes with capacities but without successors
    for (int i = 0; i < listNode.size(); i++) {
        node n;
        n.capaMem = listNode[i].capaMem;
        n.id = listNode[i].id;
        n.dis = listNode[i].dis;
        n.preNex = listNode[i].preNex;
        (*listNodeT)[i] = n;
    }

    // Reverse the arcs
    for (int i = 0; i < listNode.size(); i++) {
        for (int j = 0; j < listNode[i].succ.size(); j++) {
            tuple<node*, int> t(&(*listNodeT)[i], get<1>(listNode[i].succ[j]));
            (*listNodeT)[get<0>(listNode[i].succ[j])->id].succ.push_back(t);
        }
    }

    return listNodeT;
}



// Function to compute the auxiliary shortest paths using Eppstein's algorithm.
// The computation is performed directly on the node list.
void eppsteinF(vector<node>* listNode) {
	for (int i = 0; i < (*listNode).size(); i++) {
		for (int j = 0; j < (*listNode)[i].succ.size(); j++) {
			// The value of the arc according to Eppstein's representation is calculated as follows:
			// distance from auxiliary path - distance of shortest path + distance to reach auxiliary node
			get<1>((*listNode)[i].succ[j]) = (*listNode)[get<0>((*listNode)[i].succ[j])->id].dis - (*listNode)[i].dis + get<1>((*listNode)[i].succ[j]);
		}
	}

	for (int i = 0; i < (*listNode).size(); i++) {
		for (int j = 0; j < (*listNode)[i].succ.size(); j++) {
			// If the successor node is actually the preNex node (part of the shortest path), remove it from successors.
			if (get<0>((*listNode)[i].succ[j])->id == (*listNode)[i].preNex)
				eraseSucc(&((*listNode)[i].succ), (*listNode)[i].succ[j]);
		}
	}
}

// Function to compute auxiliary paths using a set of auxiliary nodes.
vector<tuple<int, vector<int>>> lowPathAux(vector<node> listNode, vector<int> aux, int src, int dst) {
	vector<tuple<int, vector<int>>> path; // Path to return that passes through all auxiliary nodes
	bool b = false; // Flag to indicate if the current node is in the aux set
	bool stop = false; // Flag to indicate if the path construction should stop (e.g., due to a loop)
	int v = 0; // Index to traverse auxiliary nodes, incrementing when an aux node is found
	int j = 0; 
	int i = listNode[src].id; // Represents the node id
	path.push_back({i,{}});

	// Check if the source node is in the aux array and increment v if true
	if(aux.size() > 0 && listNode[src].id == aux[v]){
		v++; 
	}

	// Loop until the destination is reached or all auxiliary nodes are visited
	while ((i >= 0) && (!stop) && (v < aux.size())) {
		j = 0;
		b = false;

		// Traverse successors to see if any is equal to aux[v]
		while ((j < listNode[i].succ.size()) && (!b)) {
			if (get<0>(listNode[i].succ[j])->id == aux[v]) {
				// Before adding to path, check if it already exists to avoid loops
				for (int r = 0; r < path.size(); r++) {
					if (get<0>(path[r]) == aux[v]) {
						stop = true;
					}
				}
				if(!stop){
					path.push_back({ aux[v],{} });
					v++; // Increment v after adding to path
					i = get<0>(listNode[i].succ[j])->id;
					b = true;
				}
			}
			j++;
		}

		// If the current node's successors don't contain aux[v], add the direct successor to the path
		if (!b) {
			i = listNode[i].preNex;
			if (i >= 0) {
				// Check if the next direct node is already in the path
				for(int r = 0; r < path.size(); r++) {
					if (get<0>(path[r]) == i) { 
						stop = true;
					}
				}
				if (!stop) {
					path.push_back({ i,{} });
					// Check if the direct successor is the next auxiliary node
					if (aux[v] == i) {
						v++;
					}
				}
			}
		}
	}

	// If all auxiliary nodes have been visited, ensure that the path ends at the destination
	if(v == aux.size()){
		while ((!stop) && (listNode[i].id != dst)) {
			i = listNode[i].preNex;
			if (i >= 0) {
				// Check for loops by ensuring the node is not already in the path
				for (int r = 0; r < path.size(); r++) {
					if (get<0>(path[r]) == i) {
						stop = true; // Stop if a loop is detected
					}
				}
				if (!stop) {
					path.push_back({ i,{} });
				}
			}
		}
	}

	// Return the constructed path if no loops were detected
	if(!stop){
		return(path);
	} else {
		// If a loop is detected, return a path with just -1 indicating an error
		vector<tuple<int, vector<int>>> pathStop;
		pathStop.push_back({ -1,{} });
		return(pathStop);
	}
}


// Returns the index of the node with the smallest distance to the destination node
int lowestWeight(vector<node>* listNodeP) {
	int lW = 9999; // Initialize with a large number to represent infinity
	int lI = -1; // Index of the node with the lowest weight
	// Iterate through the node list to find the node with the lowest weight
	for (int i = 0; i < (*listNodeP).size(); i++) {
		// If the node's distance is less than the current lowest and it has not been visited
		if (((*listNodeP)[i].dis < lW) && !(*listNodeP)[i].v) {
			lI = i;
			lW = (*listNodeP)[i].dis;
		}
	}
	// If no node is found, return -1 indicating an error
	if (lI == -1) {
		return -1;
	} else {
		// Mark the found node as visited
		(*listNodeP)[lI].v = true;
		return lI;
	}
}

// Computes the shortest path using Dijkstra's algorithm directly on the node list
void djikstra(vector<node>* listNodeP, int source, int dest, int nbr) {
	// Initialize the distance of the source node to 0
	(*listNodeP)[source].dis = 0;
	bool stop = true;
	// Iterate until all nodes have been visited
	while (stop) {
		int i = lowestWeight(listNodeP);
		if (i != -1) {
			// Relax the edges
			for (int j = 0; j < (*listNodeP)[i].succ.size(); j++) {
				if (get<0>((*listNodeP)[i].succ[j])->dis >= (*listNodeP)[i].dis + get<1>((*listNodeP)[i].succ[j])) {
					// Update the predecessor and distance
					get<0>((*listNodeP)[i].succ[j])->preNex = (*listNodeP)[i].id;
					get<0>((*listNodeP)[i].succ[j])->dis = (*listNodeP)[i].dis + get<1>((*listNodeP)[i].succ[j]);
				}
			}
		} else {
			stop = false;
		}
	}
}

// Returns the shortest path after running Dijkstra's algorithm
vector<int> lowPathF(vector<node> listNodeT, int src, int dst) {
	vector<int> lowPath; // Vector to store the lowest path
	int i = listNodeT[src].id;
	// Traverse from the source node to the destination
	do {
		lowPath.push_back(i);
		i = listNodeT[i].preNex;
	} while (i >= 0);

	return lowPath;
}

// Returns the total distance value of a path
int lowPathDisF(vector<node> listNode, vector<int> lowPath) {
	int valDis = listNode[0].dis; // Initialize with the distance of the first node
	// Iterate over the path to calculate the total distance
	for (int i = 0; i < lowPath.size() - 1; i++) {
		// If the next node is not the direct predecessor
		if (lowPath[i + 1] != listNode[lowPath[i]].preNex) {
			// Add the distance of the corresponding edge
			for (int j = 0; j < listNode[lowPath[i]].succ.size(); j++) {
				if (get<0>(listNode[lowPath[i]].succ[j])->id == lowPath[i + 1]) {
					valDis += get<1>(listNode[lowPath[i]].succ[j]);
				}
			}
		}
	}
	return valDis;
}

// Adds the child nodes of an Eppstein list node to the heap
void addChild(treeNode* orig, int val, MinHeap* h, int cosVal, vector<int> parent) {
	vector<treeNode*> noVis; // Non-visited nodes
	noVis.push_back(orig);
	// Traverse the tree until the current node equals 'val' and has the same parents
	while ((noVis[0]->idNodeGraph != val) || (noVis[0]->parent != parent)) {
		for (int i = 0; i < noVis[0]->succ.size(); i++) {
			noVis.push_back(noVis[0]->succ[i]);
		}
		// Remove the first element from the non-visited list
		eraseTreeNodeVect(&noVis, *noVis[0]);
	}
	// Verify that the current node equals 'val' and has the same parents
	if ((noVis[0]->idNodeGraph == val) && (noVis[0]->parent == parent)) {
		// Add the successors to the heap with the updated cost
		for (int j = 0; j < noVis[0]->succ.size(); j++) {
			h->insertKey({ noVis[0]->succ[j]->wei + cosVal, noVis[0]->succ[j]->idNodeGraph, noVis[0]->succ[j]->parent });
		}
	} else {
		std::cerr << "Error in function AddChild" << std::endl;
	}
}


// Returns the k-th path in the form of auxiliary nodes that compose it
tup kPath(treeNode* orig, MinHeap* h) {
    tup l = h->getMin(); // Get the minimum element from the heap
    int t = get<0>(l);  // Weight of the auxiliary path
    int tt = get<1>(l);  // ID of the auxiliary node
    vector<int> j = get<2>(l);        // Parent nodes
    h->extractMin(); // Remove the minimum element from the heap
    addChild(orig, tt, h, t, j); // Add child nodes to the heap
    return { t,tt, j };
}

// Computes the next path by considering the auxiliary nodes and the last k-th path
vector<tuple<int, vector<int>>> nextPath(treeNode* orig, MinHeap* h, vector<node> listNode, vector<tuple<int, vector<int>>> lastKPath, int src, int dst) {
    tup tp;
    vector<tuple<int, vector<int>>> path;
    bool bl = true;
    tup sauv = { -1, -1, {} };
    tp = kPath(orig, h); // Get the k-th path
    
    if (get<0>(tp) != -1) {
        // Get the low auxiliary path given the auxiliary nodes
        path = lowPathAux(listNode, get<2>(tp), src, dst);
        // If the current path is different from the last k-th path, return it
        if (path != lastKPath) {
            return path;
        } else { // If the path is the same as the previous path
            return {}; 
        }
    } else {
        return {};
    }
}

// Constructs the Eppstein tree and returns a pointer to the first node
treeNode eppsteinListNode(vector<node>* listNodeE, int src, int dst, int* treeSize) {
    bool indB;
    treeNode orig;
    treeNode* leaf = (treeNode*)malloc(sizeof(treeNode));
    
    // Create the origin node corresponding to the direct path without deviation through auxiliary nodes
    int id = 0;
    orig.id = id++;
    orig.wei = 0;
    orig.idNodeGraph = src;
    orig.parent = { src };
    
    // Create a list to visit (lisEx) and another for saving (lisExSau)
    vector<treeNode*> lisEx;
    vector<treeNode*> lisExSau;
    // Insert the first node into the list to visit
    lisEx.push_back(&orig);
    lisExSau.push_back(&orig);

    // Indicator of the node to be processed
    int idNode;
    
    int nbKPathInd = 0; // This indicator is incremented when a shortest path is added
    int nbKPath = 100; // This indicator represents the number of iterations
    
    // While there are nodes to visit and we haven't reached the number of k paths
    while ((lisEx.size() > 0) && (nbKPathInd < nbKPath)) {
        idNode = lisEx[0]->idNodeGraph;
        do {
            // Visit the successors of the current node
            for (int g = 0; g < ((*listNodeE)[idNode].succ).size(); g++) {
                indB = true; // Assume the node can be added until proven otherwise

                // Create a leaf node and add its parents
                leaf = new treeNode;
                leaf->id = id++;
                leaf->wei = get<1>((*listNodeE)[idNode].succ[g]);
                leaf->idNodeGraph = get<0>((*listNodeE)[idNode].succ[g])->id;
                leaf->parent = lisEx[0]->parent;
                leaf->parent.push_back(leaf->idNodeGraph);

                // Check for duplicates in the parent list to avoid loops
                for (int l = 0; l < leaf->parent.size(); l++) {
                    for (int ll = 0; ll < leaf->parent.size(); ll++) {
                        if ((leaf->parent[l] == leaf->parent[ll]) && (l != ll)) {
                            indB = false;
                        }
                    }
                }

                // Check if the node has been processed before
                if (indB) {
                    for (int v = 0; v < lisExSau.size(); v++) {
                        if ((lisExSau[v]->idNodeGraph == leaf->idNodeGraph) && (lisExSau[v]->parent == leaf->parent)) {
                            indB = false;
                        }
                    }
                }
                
                // Ensure the last node is not already in the previous path to avoid loops
                if (indB) {
                    int nn = lisEx[0]->idNodeGraph;
                    do {
                        if ((*listNodeE)[nn].id == leaf->idNodeGraph) {
                            indB = false;
                        }
                        nn = (*listNodeE)[nn].preNex;
                    } while (nn != -1 && nn != dst);
                }
    
                if (indB) {
                    // Add the leaf to the successor list and the nodes to visit
                    lisEx[0]->succ.push_back(leaf);
                    lisEx.push_back(leaf);
                    lisExSau.push_back(leaf);
                    nbKPathInd++;
                }
            }
            idNode = (*listNodeE)[idNode].preNex;
        } while (idNode >= 0 && idNode != dst);
        // Remove the first element from the list to visit
        eraseTreeNodeVect(&lisEx, *lisEx[0]);
    }

    *treeSize = lisExSau.size();

    return orig;
}



// Constructs the Eppstein tree and returns a pointer to the first node
treeNode* eppsteinListNode2(vector<node>* listNodeE, int src, int dst, int* treeSize) {
    bool isValid;
    treeNode orig;
    treeNode* leaf;

    // Create the origin node corresponding to the direct path without deviation by auxiliary nodes
    int id = 0;
    orig.id = id++;
    orig.wei = 0;
    orig.idNodeGraph = src;
    orig.parent = { src };

    // Create the list listEx to visit and another for backup
    vector<treeNode*> lisEx;
    vector<treeNode*> lisExSau;
    // Insert the first node into the visit list
    lisEx.push_back(&orig);
    lisExSau.push_back(&orig);

    // Indicator of the node to be treated
    int idNode;

    int nbKPathInd = 0; // This indicator is incremented when a shorter path is added
    int nbKPath = 100; // This indicator represents the number of iterations

    // While there are nodes to visit and we haven't reached the k path limit
    while ((lisEx.size() > 0) && (nbKPathInd < nbKPath)) {
        idNode = lisEx[0]->idNodeGraph;

        do {
            // Visit the neighbors
            for (int g = 0; g < ((*listNodeE)[idNode].succ).size(); g++) {
                isValid = true; // As long as there's a neighbor, assume it's valid

                // Create the leaf as well as adding parent
                leaf = new treeNode;
                leaf->id = id++;
                leaf->wei = get<1>((*listNodeE)[idNode].succ[g]);
                leaf->idNodeGraph = get<0>((*listNodeE)[idNode].succ[g])->id;

                // Copy the list of parents from the current node and add the new parent
                leaf->parent = lisEx[0]->parent;
                leaf->parent.push_back(leaf->idNodeGraph);

                // Check if there are duplicate nodes in the parent list to avoid loops
                for (size_t l = 0; l < leaf->parent.size(); l++) {
                    for (size_t ll = l + 1; ll < leaf->parent.size(); ll++) {
                        if (leaf->parent[l] == leaf->parent[ll]) {
                            isValid = false;
                        }
                    }
                }

                // Check if the node has been processed before
                if (isValid) {
                    for (auto* existingNode : lisExSau) {
                        if (existingNode->idNodeGraph == leaf->idNodeGraph && existingNode->parent == leaf->parent) {
                            isValid = false;
                        }
                    }
                }

                // Ensure the new node is not part of the direct path which would create a loop
                if (isValid) {
                    int currentNodeId = lisEx[0]->idNodeGraph;
                    while (currentNodeId != -1) {
                        if ((*listNodeE)[currentNodeId].id == leaf->idNodeGraph) {
                            isValid = false;
                            break;
                        }
                        currentNodeId = (*listNodeE)[currentNodeId].preNex;
                    }
                }

                // If the node is valid, add it to the lists
                if (isValid) {
                    lisEx[0]->succ.push_back(leaf);
                    lisEx.push_back(leaf);
                    lisExSau.push_back(leaf);
                    nbKPathInd++;
                }
            }
            // Move to the next node on the direct path
            idNode = (*listNodeE)[idNode].preNex;
        } while (idNode >= 0 && idNode != dst);

        // Remove the first element from the visit list
        eraseTreeNodeVect(&lisEx, *lisEx[0]);
    }

    *treeSize = lisExSau.size();

    return &orig;
}

