// A C++ program to demonstrate common Binary Heap Operations
#include<iostream>
#include<climits>
#include<tuple>
using namespace std;

typedef tuple<int, int, vector<int>> tup;
//Structure representing the nodes of the node tree according to Eppstein's representation

//Structure representing the nodes of the node tree according to Eppstein's representation
struct treeNode {
	int id = -1;
	int idNodeGraph = -1; //Id of the node in the graph
	int wei = -1; //Weight if the node is selected (auxiliary nodes)
	bool v = true; //How to find out if the node has already been considered in the calculation of the k shortest paths 
	vector<treeNode*> succ; //The successors of a node in the Eppstein representation tree 
	vector<int> parent; //Reference of parent nodes, which represents the list of auxiliary nodes forming the path 

	vector<tuple<int,int>> parent2; //Reference of parent nodes, which represents the list of auxiliary nodes forming the path 
};

// Prototype of a utility function to swap two integers
void swap(tup* x, tup* y);

// A class for Min Heap
class MinHeap
{
	tup* harr; // pointer to array of elements in heap
	int capacity; // maximum possible size of min heap
	int heap_size; // Current number of elements in min heap
public:
	// Constructor
	MinHeap(int capacity);

	// to heapify a subtree with the root at given index
	void MinHeapify(int);

	int parent(int i) { return (i - 1) / 2; }

	// to get index of left child of node at index i
	int left(int i) { return (2 * i + 1); }

	// to get index of right child of node at index i
	int right(int i) { return (2 * i + 2); }

	// to extract the root which is the minimum element
	tup extractMin();

	// Decreases key value of key at index i to new_val
	void decreaseKey(int i, tup new_val);

	// Returns the minimum key (key at root) from min heap
	tup getMin() { return harr[0]; }

	// Deletes a key stored at index i
	void deleteKey(int i);

	// Inserts a new key 'k'
	void insertKey(tup k);
};

// Constructor: Builds a heap from a given array a[] of given size
MinHeap::MinHeap(int cap)
{
	heap_size = 0;
	capacity = cap;
	harr = new tup[cap];
}

// Inserts a new key 'k'
void MinHeap::insertKey(tup k)
{
	if (heap_size == capacity)
	{
		cout << "\nOverflow: Could not insertKey\n";
		return;
	}

	// First insert the new key at the end
	heap_size++;
	int i = heap_size - 1;
	harr[i] = k;

	// Fix the min heap property if it is violated
	while (i != 0 && get<0>(harr[parent(i)]) > get<0>(harr[i]))
	{
		swap(&harr[i], &harr[parent(i)]);
		i = parent(i);
	}
}

// Decreases value of key at index 'i' to new_val. It is assumed that
// new_val is smaller than harr[i].
void MinHeap::decreaseKey(int i, tup new_val)
{
	harr[i] = new_val;
	while (i != 0 && get<0>(harr[parent(i)]) > get<0>(harr[i]))
	{
		swap(&harr[i], &harr[parent(i)]);
		i = parent(i);
	}
}

// Method to remove minimum element (or root) from min heap
tup MinHeap::extractMin()
{
	tup t = { -1,-1,{} };
	if (heap_size <= 0)
		return t;
	if (heap_size == 1)
	{
		heap_size--;
		return harr[0];
	}

	// Store the minimum value, and remove it from heap
	tup root = harr[0];
	harr[0] = harr[heap_size - 1];
	heap_size--;
	MinHeapify(0);

	return root;
}


// This function deletes key at index i. It first reduced value to minus
// infinite, then calls extractMin()
void MinHeap::deleteKey(int i)
{
	tup t = { -1,-1, {} };
	decreaseKey(i, t);
	extractMin();
}

// A recursive method to heapify a subtree with the root at given index
// This method assumes that the subtrees are already heapified
void MinHeap::MinHeapify(int i)
{
	int l = left(i);
	int r = right(i);
	int smallest = i;
	if (l < heap_size && get<0>(harr[l]) < get<0>(harr[i]))
		smallest = l;
	if (r < heap_size && get<0>(harr[r]) < get<0>(harr[smallest]))
		smallest = r;
	if (smallest != i)
	{
		swap(&harr[i], &harr[smallest]);
		MinHeapify(smallest);
	}
}

// A utility function to swap two elements
void swap(tup* x, tup* y)
{
	tup temp = *x;
	*x = *y;
	*y = temp;
}


