#pragma once
#include<iostream>
#include<climits>
#include<tuple>
#include<string>
#include<vector>
#include <algorithm>

// Deploys an SFC without parallelization and returns the deployment value <node, micro-service>
vector<tuple<int, vector<int>, vector<vector<int>>>> deploySfcPath(vector<tuple<int, vector<int>>> path, vector<tuple<int, vector<int>>> SFC, vector<node>* nodeList) {
	vector<tuple<int, vector<int>, vector<vector<int>>>> deployment;
	int m = 0; // Iterates over the SFC
	int iT = 0; // Iterates over the nodes
	bool deployed = false;
	bool endOfPath = true; // Indicates if we reached the end of the path (false) or not (true)

	// Calculate the path margin = Available space - size of the SFC.
	int pathMargin = -1;
	int memoryPath = 0;
	int lengthSFC = SFC.size();

	// Calculate available capacity on a path
	for (int i = 0; i < path.size(); i++) {
		memoryPath += (*nodeList)[get<0>(path[i])].capaMem;
	}

	pathMargin = memoryPath - lengthSFC;

	if (pathMargin >= 0) {
		if (get<0>(path[0]) != -1) {
			while ((m < SFC.size()) && endOfPath) { // While the SFC has not been completely deployed and we have not reached the end of the path
				deployed = false;
				while ((iT < path.size()) && !deployed) {
					if ((*nodeList)[get<0>(path[iT])].capaMem > 0) { // Deploy micro-service m on node iT
						bool found = false;
						int idPlacementM = 0;
						while ((idPlacementM < deployment.size()) && !found) { // If node iT already exists in deployment, reference its id and deploy m
							if (get<0>(deployment[idPlacementM]) == get<0>(path[iT])) {
								get<1>(deployment[idPlacementM]).push_back(get<0>(SFC[m]));
								found = true;
							}
							if (!found) {
								idPlacementM++;
							}
						}
						if (!found) { // If iT does not exist in deployment, add it to the deployment and reference it in the variable idPlacementM
							deployment.push_back({ get<0>(path[iT]), {get<0>(SFC[m])}, {} });
							idPlacementM = deployment.size() - 1;
						}
						(*nodeList)[get<0>(path[iT])].capaMem--;
						deployed = true;
						m++;
					}
					else {
						iT++;
					}
				}
				if (iT >= path.size()) {
					endOfPath = false;
				}
			}
		}
	}

	// If we exit the loop with endOfPath = false, erase the deployment
	if (!endOfPath) {
		for (int j = 0; j < deployment.size(); j++) {
			for (int t = 0; t < get<1>(deployment[j]).size(); t++) {
				(*nodeList)[get<0>(deployment[j])].capaMem++; // If deployment was made, update the memory capacities
			}
		}
		return {};
	}
	return deployment;
}

// Removes a micro-service from a set of micro-services
vector<int> removeMicroService(int m, vector<int> set) {
	vector<int> result;
	for (int i : set) {
		if (i != m) {
			result.push_back(i);
		}
	}
	return result;
}

// Indicates if two micro-services function in parallel or not.
bool areParallel(vector<tuple<int, vector<int>>> q, int m, int m2) {
	for (auto& elem : q) {
		if (get<0>(elem) == m2) {
			for (int val : get<1>(elem)) {
				if (m == val) {
					return false;
				}
			}
		}
	}
	return true;
}

// Determines if a micro-service m belongs to a parallelism set contained in node n
int belongsTo(int m, vector<vector<int>> sets) {
	for (int i = 0; i < sets.size(); i++) {
		for (int val : sets[i]) {
			if (val == m) {
				return i;
			}
		}
	}
	return -1;
}

// Deploys an SFC by parallelizing during the deployment and returns the deployment value <node, micro-service>
vector<tuple<int, vector<int>, vector<vector<int>>>> deploySfcPathWithInternalParallelization(vector<tuple<int, vector<int>>> path, vector<tuple<int, vector<int>>> SFC, int latency, vector<node>* nodeList, int idSFC, int depEqui) {
	// Check if deployment is possible on the path
	// Create an array of vectors of the same size as the number of nodes in the shortest path
	// For each micro-service of the SFC
	// t = 0
	// While t < size of SFC and indexTab < size of the array
	// If the capacity of node indexTab allows
		// Update the array
		// indexTab++
	// Otherwise
		// Move to the next node: indexTab++
	vector<tuple<int, vector<int>, vector<vector<int>>>> deployment;
	int m = 0; // Iterates over the SFC
	int iT = 0; // Iterates over the nodes
	bool deployed = false;
	bool endOfPath = true; // Indicates if we have reached the end of the path (false) or not (true)

	// Calculate the path margin = Available space - size of the SFC.
	int pathMargin = -1;
	int DS = 0;
	int lengthSFC = SFC.size();

	// Calculate available capacity on a path
	for (auto& p : path) {
		DS += (*nodeList)[get<0>(p)].capaMem;
	}

	pathMargin = DS - lengthSFC;

	// Calculate variables necessary for load-balanced deployment
	int nodeMargin = pathMargin / SFC.size(); // Corresponds to the integer part of the margin divided by the number of nodes
	double residualMargin = static_cast<double>(pathMargin) / SFC.size() - nodeMargin;
	int balanceParameter = nodeMargin; // If balanced deployment, then this quantity should remain on each node, otherwise set to zero for unbalanced deployment
	if (!depEqui) {
		balanceParameter = 0;
	}
	bool criticalitySFC = true; // Defines the criticality of an SFC (if parallelization necessary or not)
	// Calculate the criticalitySFC indicator
	int arcLatency = 1; // Corresponds to the latency of the arcs here at 1ms
	int executionDuration = 1; // Corresponds to the latency of execution of the micro-services
	if (((path.size() - 1) * arcLatency) + (SFC.size() * (executionDuration)) <= latency) { // If the latency on the arcs + execution latency is less than the required latency, then SFC is not critical
		criticalitySFC = false;
	}


	if (margeSFC >= 0) {
		if (get<0>(path[0]) != -1) {
			while ((m < SFC.size()) && (depSI)) { // If SFC is not fully deployed (visited) and we have not reached the end of the path and the path is not composed of -1 only
				bool depI = false; // Indicates whether S m has been deployed or not
				while ((iT < path.size()) && (!depI)) {
					// if ((*nodeList)[get<0>(path[iT])].capaMem > 0) { // Deploy S m on node iT
					if ((*nodeList)[get<0>(path[iT])].capaMem > parametreEquilibre) { // New version: deploy until the limit corresponding to the margin per node is reached
						bool ind = false; // Indicates if we were able to find the current node in the deployment
						int idPlacementM = 0;
						while ((idPlacementM < dep.size()) && (!ind)) { // If node iT already exists in the deployment, reference its id and deploy m
							if (get<0>(dep[idPlacementM]) == get<0>(path[iT])) {
								get<1>(dep[idPlacementM]).push_back(get<0>(SFC[m]));
								ind = true;
							}
							if (!ind) {
								idPlacementM++;
							}
						}
						if (!ind) { // If iT does not exist in the deployment, add it to the deployment and reference it in idPlacementM
							dep.push_back({ get<0>(path[iT]), {get<0>(SFC[m])}, {} });
							idPlacementM = 0;
							while ((idPlacementM < dep.size()) && (!ind)) {
								if (get<0>(dep[idPlacementM]) == get<0>(path[iT])) {
									ind = true;
								}
								if (!ind) {
									idPlacementM++;
								}
							}
						}
						// dep.push_back({path[iT], get<0>(SFC[m])});
						(*nodeList)[get<0>(path[iT])].capaMem--;
						depI = true;
						// DO THIS ONLY IF THE PATH SUM + EXECUTION LATENCY OF ALL S > REQUIRED LATENCY
						// First, check if the previous micro-service m is parallelizable with m
						// Check if it is placed on the same node
						// If yes ==> Parallelize the micro-services
						// Otherwise, check if margeSFC > 0
						// If yes, check if it works in parallel
						// If it does not work in parallel, check if there is space on the node iT
						// If there is space, move the S on the node iT and reduce the margin by 1
						// If there is no space, stop
						// If it works in parallel, check Stop. For now 
						bool placedSameNode = false; // To block parallelization
						// int nodePlacementIndex = -1;
						if (para(SFC, get<0>(SFC[(m - 1)]), get<0>(SFC[m])) && (m > 0)) { // Check if the two micro-services are parallelizable and not on the first S of the SFC
							for (int g = 0; g < dep.size(); g++) { // Check if m-1 is placed on the same node
								if (get<0>(dep[g]) == get<0>(path[iT])) {
									for (int l = 0; l < get<1>(dep[g]).size(); l++) {
										if (get<1>(dep[g])[l] == get<0>(SFC[m - 1])) {
											placedSameNode = true;
										}
									}
								}
							}
							if (placedSameNode) { // If placed on the same node ==> No imbalance OK
								// Check if it does not already work in parallel
								int indiEnsPara = -2;
								indiEnsPara = appartient(get<0>(SFC[m - 1]), get<2>(dep[idPlacementM]));
								if (indiEnsPara == -1) { // If it does not work in parallel, place it in the current node and
									// Place m in a new parallelism group
									vector<int> para;
									para.push_back(get<0>(SFC[m]));
									// Place m-1 in the same group
									para.push_back(get<0>(SFC[m - 1]));
									get<2>(dep[idPlacementM]).push_back(para);
								}
								else { // If it already works in parallel, check if the current micro-service can also work with all the micro-services
									bool indiParaTT = true;
									for (int u = 0; u < get<2>(dep[idPlacementM])[indiEnsPara].size(); u++) {
										if (!para(SFC, get<2>(dep[idPlacementM])[indiEnsPara][u], get<0>(SFC[m]))) {
											indiParaTT = false;  // If it is not parallelizable with at least one micro-service
										}
									}
									if (indiParaTT) { // If m is parallelizable with all the micro-services in the group
										get<2>(dep[idPlacementM])[indiEnsPara].push_back(get<0>(SFC[m]));
									}
									else {
										// m is not parallelizable with at least one S, so do nothing
									}
								}
							}
							else { // If the two S are not placed on the same node ==> Implies imbalance ==> check if SFC is critical
								int indiNoeudPrecM = -1;
								if (margeSFC > 0) { // Check if the margin is greater than 0 and the SFC is critical
									// Check if m-1 does not already work in parallel
									for (int g = 0; g < dep.size(); g++) { // Retrieve the node where m-1 is placed
										for (int l = 0; l < get<1>(dep[g]).size(); l++) {
											if (get<1>(dep[g])[l] == get<0>(SFC[m - 1])) { // Once we find the micro-service, we retrieve the node indicator
												indiNoeudPrecM = g;
											}
										}
									}
									int indiGrParaPreM = -2;
									indiGrParaPreM = appartient(get<0>(SFC[m - 1]), get<2>(dep[indiNoeudPrecM])); // Check if m-1 does not already work in parallel
									if ((indiGrParaPreM == -1)) { // if it does not work in parallel, create a para group for m and m-1
										if ((margeResid > 0) || criticiteSFC) { // And also if the RSS (residual margin) > 0 or if SFC is critical (the margin has already been checked)
											vector<int> para;
											// Place m in the para group
											para.push_back(get<0>(SFC[m]));
											// Place m-1 in the same group
											para.push_back(get<0>(SFC[m - 1]));
											// Place the group in the corresponding node in the deployment
											get<2>(dep[idPlacementM]).push_back(para);
											// Remove m-1 from the node indiNoeudPrecM
											get<1>(dep[indiNoeudPrecM]) = removeMS(get<0>(SFC[m - 1]), get<1>(dep[indiNoeudPrecM]));
											// Free up space on the node where m-1 was placed
											(*nodeList)[get<0>(dep[indiNoeudPrecM])].capaMem++; // To be checked... and add ++ for the following node...
											// Place m-1 on the node idPlacementM
											get<1>(dep[idPlacementM]).push_back(get<0>(SFC[m - 1]));
											// Decrease space on the node where m is placed
											(*nodeList)[get<0>(dep[idPlacementM])].capaMem--;
										}
									}
									else {
										// If m-1 works in parallel, no action needed.
										// Consider checking if the margin is large enough to move everything to the iT node
										// Retrieve the index of the node where S m is placed

										// If the SFC margin and the size of the iT node allow the movement of all parallelizable S
										if ((margeSFC >= get<2>(dep[indiNoeudPrecM])[indiGrParaPreM].size()) && ((*nodeLis)[get<0>(path[iT])].capaMem >= get<2>(dep[indiNoeudPrecM])[indiGrParaPreM].size())) {
											// Check if the residual margin is sufficient for the microservice movement, or if the SFC is critical
											if ((margeResid >= get<2>(dep[indiNoeudPrecM])[indiGrParaPreM].size()) || criticiteSFC) {
												// Check if S m can work with the parallelizables of m-1
												bool indiParaTT = true;
												for (int u = 0; u < get<2>(dep[indiNoeudPrecM])[indiGrParaPreM].size(); u++) {
													if (!para(SFC, get<2>(dep[indiNoeudPrecM])[indiGrParaPreM][u], get<0>(SFC[m]))) {
														indiParaTT = false;  // If it is not parallelizable with at least one micro-service
													}
												}
												if (indiParaTT) { // If m is parallelizable with all the parallelizables of m-1
													// Create a set of parallelizables containing m, m-1, and other parallelizables
													vector<int> paraEns;
													// Add each parallelizable S and move them simultaneously to the iT node
													int u = 0;
													while (get<2>(dep[indiNoeudPrecM])[indiGrParaPreM].size() > 0) {
														// Add to the parallelizable set
														paraEns.push_back(get<2>(dep[indiNoeudPrecM])[indiGrParaPreM][u]);
														// Deploy on the iT node
														get<1>(path[iT]).push_back(get<2>(dep[indiNoeudPrecM])[indiGrParaPreM][u]);
														(*nodeLis)[get<0>(path[iT])].capaMem--;
														// Remove from the initial node
														get<1>(dep[indiNoeudPrecM]) = removeMS(get<2>(dep[indiNoeudPrecM])[indiGrParaPreM][u], get<1>(dep[indiNoeudPrecM]));
														(*nodeLis)[get<0>(dep[indiNoeudPrecM])].capaMem++;
														// Remove the micro-service from the initial parallelizable set
														get<2>(dep[indiNoeudPrecM])[indiGrParaPreM] = removeMS(get<2>(dep[indiNoeudPrecM])[indiGrParaPreM][u], get<2>(dep[indiNoeudPrecM])[indiGrParaPreM]);
													}
													// Add m to paraEns
													paraEns.push_back(get<0>(SFC[m]));
													// Add the paraEns set to the iT node
													// Retrieve the index of the iT node in dep
													int indiceNoeudPlacementM = -1;
													for (int g = 0; g < dep.size(); g++) {
														if (get<0>(dep[g]) == get<0>(path[iT])) {
															indiceNoeudPlacementM = g;
														}
													}
													get<2>(dep[indiceNoeudPlacementM]).push_back(paraEns);
												}
											}
											// Do nothing if m is not parallelizable with at least one parallelizable of m-1
										}
									}
									// To block the parallelization press here
									m++;
								}
								else {
									iT++;
								}
							}
							if (iT >= path.size()) {
								depSI = false;
							}
						}
					}
				}



				// Exiting the loop with depSI as false indicates deployment should be overwritten
				// This means either the deployment is not complete
				// or the path for deployment contains a single node with value -1
				if (depSI == false) {
					for (int j = 0; j < dep.size(); j++) {
						for (int t = 0; t < get<1>(dep[j]).size(); t++) {
							(*nodeLis)[get<0>(dep[j])].capaMem++; // Update memory capacities if deployment is realized
						}
					}
					vector<tuple<int, vector<int>, vector<vector<int>>>> depNull;
					return depNull;
				}
				return dep;
			}

			// Creates chaining (adding successors for each node) for a path
			// This chaining must not include bifurcation or fusion
			void chaining(vector<tuple<int, vector<int>>>*kPath) {
				for (int i = 0; i < (*kPath).size() - 1; i++) {
					get<1>((*kPath)[i]).push_back(get<0>((*kPath)[i + 1]));
				}
			}

			// Comparison function for ordering sets of scoreSFC
			bool minMax(const scoreSFC & a, const scoreSFC & b) {
				return a.score < b.score;
			}
			bool maxMin(const scoreSFC & a, const scoreSFC & b) {
				return a.score > b.score;
			}

			void permute(vector<int>&nums, int index, vector<vector<int>>&result) {
				if (index == nums.size() - 1) {
					result.push_back(nums);
					return;
				}
				for (int i = index; i < nums.size(); i++) {
					swap(nums[index], nums[i]);
					permute(nums, index + 1, result);
					swap(nums[index], nums[i]);
				}
			}

			vector<vector<int>> permutations(vector<int>&nums) {
				vector<vector<int>> result;
				permute(nums, 0, result);
				return result;
			}




			// Attempt a deployment for each Service Function Chain (SFC)
			vector<deployment> deployementS(vector<scoreSFC> orderedSFC, vector<node>*listNode, int in, int depEqui, int paraInt) {
				vector<deployment> resDep; // Stores the deployment results
				bool notDep;               // Flag to indicate if deployment has not been done
				bool solution;             // Flag to indicate if a solution is possible
				vector<node> listNodeE;    // A list to store nodes for processing
				vector<int> lowPath;       // To store the shortest path
				int sizeEppsteinTree = 0;  // The size of the Eppstein tree

				int src, dst;              // Source and destination for paths
				vector<tuple<int, vector<int>>> kPath; // k-th path storage

				// Code for deployment procedure continues...
				// The details of the deployment algorithm are not shown here for brevity
			}

			// The commented out code contains previous versions or alternative strategies for ordering
			// the SFCs based on their margins or other criteria. It seems these have been left in as notes
			// for potential future reference or as part of the code's evolution history.

			// V1: Order from smallest to largest
			// V2: Learn from the deployment
			// Add a deployment order table

			/* Creation of an ordered set according to a defined score
			vector<scoreSFC> orderedSFC;

			// V0: Calculate the margin of each SFC and save it in the set intermOrdredSFC
			for (int q = 0; q < Q.size(); q++) {
			// Calculate the margin of each SFC
			// Margin(SFC) = required latency - execution latency = required latency - length of the SFC
			int marge = latenceDemande[q] - Q[q].size();
			scoreSFC sf;
			sf.SFC = Q[q];
			sf.score = marge;
			orderedSFC.push_back(sf);
			}

			// Order the SFCs according to the set of possible permutations
			for (int q = 0; q < Q.size(); q++) {
			// Calculate the margin of each SFC
			// Margin(SFC) = required latency - execution latency = required latency - length of the SFC
			int marge = latenceDemande[q] - Q[q].size();
			scoreSFC sf;
			sf.SFC = Q[q];
			sf.score = permutationsSet[in][q];
			orderedSFC.push_back(sf);
			}

			// V1: Order the set in ascending order of margins i.e., SFCs with the tightest margins
			// are introduced and processed first in the set
			sort(orderedSFC.begin(), orderedSFC.end(), minMax);
			*/


			// Attempt to deploy for each Service Function Chain (SFC)
			vector<deployment> deployementS(vector<scoreSFC> orderedSFC, vector<node>*listNode, int in, int depEqui, int paraInt) {
				vector<deployment> resDep; // Resultant deployments

				// Variables to keep track of source, destination, and graph inversion
				int src, dst;
				vector<node> listNodeE;
				vector<int> lowPath; // To store the lowest path
				int sizeEppsteinTree = 0; // Size of the Eppstein tree for path calculations

				// Loop through each SFC to calculate the shortest path and perform deployment
				for (int q = 0; q < (orderedSFC).size(); q++) {
					src = (orderedSFC)[q].src;
					dst = (orderedSFC)[q].dst;

					// Invert the graph for reverse path calculations
					listNodeE = graphInverse(*listNode);

					// Compute the shortest path from DST to SRC using Dijkstra's algorithm
					djikstra(&listNodeE, dst, src, 0);

					// Calculate the shortest path from SRC to DST
					lowPath = lowPathF(listNodeE, src, dst);

					// Proceed only if there is a valid path from source to destination
					if (lowPath[lowPath.size() - 1] == dst) {
						// Invert the paths again for correct direction
						listNodeE = graphInverse(listNodeE);

						// Generate the Eppstein graph representation for multiple shortest paths
						eppsteinF(&listNodeE);

						// Initialize the Eppstein tree
						treeNode orig;
						vector<treeNode*> sauvv;
						orig = eppsteinListNode(&listNodeE, src, dst, &sizeEppsteinTree);

						// Initialize the min heap for Eppstein's algorithm
						MinHeap h(999);
						initMinHeap(&h, &orig);

						// Deployment related variables
						bool notDep = true; // Indicates if deployment has not been completed
						bool solution = true; // Indicates if a solution is possible
						vector<tuple<int, vector<int>>> kPath; // Stores the k-th path
						int nPath = 0; // Number of paths tried

						// Try to deploy while a solution is possible and not yet deployed
						while (notDep && solution) {
							// Get the next path from the Eppstein tree
							kPath = nextPath(&orig, &h, listNodeE, kPath, src, dst);
							nPath++;

							// Attempt to deploy the SFC on the path
							vector<tuple<int, vector<int>, vector<vector<int>>>> deploymentR;
							if (paraInt) {
								deploymentR = deplSFcPathParaInt(kPath, (orderedSFC)[q].SFC, (orderedSFC)[q].lat, listNode, (orderedSFC)[q].id, depEqui);
							}
							else {
								deploymentR = deplSFcPath(kPath, (orderedSFC)[q].SFC, listNode);
							}

							// Check if deployment was successful
							if (!deploymentR.empty()) {
								// Update deployment path and store the result
								chaining(&kPath);
								notDep = false; // Deployment completed
								resDep.push_back({ kPath, deploymentR }); // Store successful deployment
							}

							// Check if all paths have been tried without success
							if (nPath >= sizeEppsteinTree && notDep) {
								solution = false; // No more solutions possible
								resDep.push_back({}); // Store failed deployment
							}
						}
					}
					else {
						// If no valid path exists, store a failed deployment
						resDep.push_back({});
					}
				}

				return resDep; // Return the results of the deployments
			}
			// Function to calculate the number of SFCs deployed
			int deployedSFC(vector<deployment> resDep) {
				int count = 0;
				// Loop through all deployments and count the ones that have been placed
				for (int q = 0; q < resDep.size(); q++) {
					if (resDep[q].placement.size() > 0) {
						count++;
					}
				}
				return count;
			}

			// Function to remove parallelizable nodes until the path size is reached, or return 0
			tuple<vector<tuple<int, vector<int>>>, vector<int>> paraSFC(vector<tuple<int, vector<int>>> q, int taille) {
				vector<tuple<int, vector<int>>> qN; // Stores the new SFC without parallelizable nodes
				vector<int> para; // Stores nodes that are parallelizable
				bool isParallelizable;
				bool isTailleSufficient = true;
				int m = 0;

				// Loop to determine if nodes are parallelizable and remove them if necessary
				while ((m < q.size()) && (isTailleSufficient)) {
					isParallelizable = true;
					// Check if the next service in the chain can be parallelized with the current one
					for (int s = 0; s < get<1>(q[m + 1]).size(); s++) {
						if (get<1>(q[m + 1])[s] == get<0>(q[m])) {
							isParallelizable = false;
						}
					}
					// If the service is parallelizable, add it to the list
					if (isParallelizable) {
						para.push_back(get<0>(q[m + 1]));
					}
					m++;
					// If the number of non-parallelizable services reaches the desired size, stop the loop
					if (q.size() - para.size() <= taille) {
						isTailleSufficient = false;
					}
				}
				// Reconstruct the SFC without the parallelizable services
				for (int m = 0; m < q.size(); m++) {
					isParallelizable = true;
					for (int j = 0; j < para.size(); j++) {
						if (get<0>(q[m]) == para[j]) {
							isParallelizable = false;
						}
					}
					if (isParallelizable) {
						qN.push_back(q[m]);
					}
				}

				return { qN, para };
			}

			// Function to compare and return the path with higher capacity
			vector<int> highCapa(vector<int> p1, vector<int> p2, vector<node> listNode) {
				int c1 = 0;
				int c2 = 0;
				// Calculate the capacity for path 1
				for (int i = 0; i < p1.size(); i++) {
					c1 += listNode[p1[i]].capaMem;
				}
				// Calculate the capacity for path 2
				for (int i = 0; i < p2.size(); i++) {
					c2 += listNode[p2[i]].capaMem;
				}
				// Return the path with higher capacity
				return (c1 >= c2) ? p1 : p2;
			}

			// Function to calculate the total capacity of a given path
			int capa(vector<int> p, vector<node> listNode) {
				int totalCapacity = 0;
				for (int i = 0; i < p.size(); i++) {
					totalCapacity += listNode[p[i]].capaMem;
				}
				return totalCapacity;
			}

			// Function to check if all predecessors of m2 are located before m in the chain
			bool precedAvant(vector<tuple<int, vector<int>>> q, int m, int m2) {
				int indexM = -1;
				int indexM2 = -1; // Indices of the two services in the chain

				// Find indices of m and m2 in the chain
				for (int i = 0; i < q.size(); i++) {
					if (get<0>(q[i]) == m) {
						indexM = i;
					}
					if (get<0>(q[i]) == m2) {
						indexM2 = i;
					}
				}

				// Check if all predecessors of m2 are located before m
				for (int p = 0; p < get<1>(q[indexM2]).size(); p++) {
					for (int j = 0; j < q.size(); j++) {
						if (get<0>(q[j]) == get<1>(q[indexM2])[p]) {
							if (j > indexM) {
								return false; // m2 is located after m, hence not all predecessors are before m
							}
						}
					}
				}

				return true; // All predecessors of m2 are before m
			}

			// Function to check if a microservice 'm' is parallelizable with all microservices in the set 'ensPr'
			bool paraEnsm(vector<int> ensPr, int m, vector<tuple<int, vector<int>>> q) {
				// Iterate over all microservices in 'ensPr' to check for parallelizability with 'm'
				for (int i = 0; i < ensPr.size(); i++) {
					if (!para(q, ensPr[i], m)) {
						return false; // If any is not parallelizable, return false
					}
				}
				return true; // All microservices in 'ensPr' are parallelizable with 'm'
			}

			// Function to create sets of microservices operating in parallel for a deployment 'qDep' and an SFC 'q'
			void paraInterF(deployment * qDep, vector<node> listNode, vector<tuple<int, vector<int>>> q) {
				// Iterate through nodes in the deployment
				for (int n = 0; n < qDep->placement.size(); n++) {
					// Check if there are microservices placed on the node
					if (get<1>(qDep->placement[n]).size() > 0) {
						// Iterate through microservices placed on the node
						for (int m = 0; m < get<1>(qDep->placement[n]).size() - 1; m++) {
							// Check if microservice 'm' does not belong to any parallelism set
							if (appartient(get<1>(qDep->placement[n])[m], get<2>(qDep->placement[n])) == -1) {
								// Iterate through other microservices placed on the same node
								for (int m2 = m + 1; m2 < get<1>(qDep->placement[n]).size(); m2++) {
									// Check if microservices 'm' and 'm2' are parallelizable
									if (para(q, get<1>(qDep->placement[n])[m], get<1>(qDep->placement[n])[m2])) {
										// Check if 'm2' does not belong to any parallelism set
										if (appartient(get<1>(qDep->placement[n])[m2], get<2>(qDep->placement[n])) == -1) {
											// Check if predecessors of 'm2' are located before 'm'
											if (precedAvant(q, get<1>(qDep->placement[n])[m], get<1>(qDep->placement[n])[m2])) {
												// Check if 'm' belongs to a parallelism set
												int h = appartient(get<1>(qDep->placement[n])[m], get<2>(qDep->placement[n]));
												if (h != -1) {
													// If 'm2' is parallelizable with all microservices in the set 'h'
													if (paraEnsm(get<2>(qDep->placement[n])[h], get<1>(qDep->placement[n])[m2], q)) {
														// Add 'm2' to the parallelism set 'h'
														get<2>(qDep->placement[n])[h].push_back(get<1>(qDep->placement[n])[m2]);
														cout << "Parallelization 1!" << endl;
													}
												}
												else {
													// Create a new parallelism set and add it to the deployment
													vector<int> paraEn;
													paraEn.push_back(get<1>(qDep->placement[n])[m]);
													paraEn.push_back(get<1>(qDep->placement[n])[m2]);
													get<2>(qDep->placement[n]).push_back(paraEn);
													cout << "Parallelization 2!" << endl;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}

			// Function to return the latency of arcs for a given path
			int latArcF(vector<node> listNode, vector<tuple<int, vector<int>>> path) {
				int latArc = 0;
				// If the path is empty, return 0
				if (path.size() <= 0) {
					return 0;
				}
				// Calculate the latency of arcs in the path
				for (int i = 0; i < path.size() - 1; i++) {
					for (int j = 0; j < listNode[get<0>(path[i])].succ.size(); j++) {
						if (get<0>(listNode[get<0>(path[i])].succ[j])->id == listNode[get<0>(path[i + 1])].id) {
							latArc += get<1>(listNode[get<0>(path[i])].succ[j]);
						}
					}
				}
				return latArc; // Return the total latency
			}

			// Returns the execution latency of microservices for a given deployment
			int latEx(vector<tuple<int, vector<int>, vector<vector<int>>>> resDep) {
				int latEx = -1;  // Initialize with -1 to indicate no latency calculated yet
				int latS = 0;    // Latency of the parallel set
				int sumLat = 0;  // Sum of latencies

				// Check if there is at least one node where microservices are deployed
				if (!resDep.empty()) {
					latEx = 0;
					for (const auto& node : resDep) {  // Iterate through the nodes
						latEx += get<1>(node).size();  // Add the deployment latency of microservices

						// Iterate through the sets of parallelism
						for (const auto& parallelSet : get<2>(node)) {
							latS = 0;  // Reset latency for the set
							sumLat = static_cast<int>(parallelSet.size());
							// Update the final latency for the node by considering parallel execution
							latEx = latEx - sumLat + 1;
						}
					}
					return latEx;
				}
				return latEx;  // Return -1 if no deployments
			}

			// Returns the latency of a service function chain (SFC)
			int latSFC(vector<node> listNode, vector<tuple<int, vector<int>>> path, vector<tuple<int, vector<int>, vector<vector<int>>>> resDep) {
				int arcLatency = latArcF(listNode, path);  // Latency of arcs
				int execLatency = latEx(resDep);           // Execution latency
				return (arcLatency + execLatency);         // Total latency is the sum of arc and execution latencies
			}

			// Calculates the cumulative latency and returns the number of SFCs deployed within the required latency
			int SFCresLat(vector<node> listNode, vector<deployment> resDep, vector<scoreSFC> orderedSFC, int* sumLat) {
				*sumLat = 0;  // Initialize the sum of latencies
				int countWithinLatency = 0;  // Counter for SFCs within the required latency

				// Iterate through the deployments
				for (int q = 0; q < resDep.size(); q++) {
					int ml = latSFC(listNode, resDep[q].chain, resDep[q].placement);
					if (ml > 0) {
						*sumLat += ml;  // Update the cumulative latency
						// If the latency is within the required limit, increment the counter
						if (ml <= orderedSFC[q].lat) {
							countWithinLatency++;
						}
					}
				}
				return countWithinLatency;  // Return the number of SFCs within the required latency
			}

			// Returns the sum of squared differences between the latencies and the average latency
			double sommeLatMinusMoy(vector<node> listNode, vector<deployment> resDep, int moy) {
				double sum = 0.0;
				// Iterate through the deployments
				for (const auto& dep : resDep) {
					int ml = latSFC(listNode, dep.chain, dep.placement);
					if (ml > 0) {
						sum += pow((ml - moy), 2);
					}
				}
				return sum;
			}

			// Function to assign a score to each SFC which is the margin between required latency and execution latency
			vector<scoreSFC> attribuerScoreMarge(vector<vector<tuple<int, vector<int>>>> Q, vector<int> latenceDemande, vector<int> sources, vector<int> destinations) {
				vector<scoreSFC> orderedSFC;
				// Iterate through the list of SFCs
				for (int q = 0; q < Q.size(); q++) {
					scoreSFC sf;
					// Set the source, destination, and latency
					sf.src = sources[q];
					sf.dst = destinations[q];
					sf.lat = latenceDemande[q];
					sf.id = q;

					// Calculate the margin for each SFC
					int marge = latenceDemande[q] - static_cast<int>(Q[q].size());
					// Detect parallel running SFCs and adjust the margin accordingly
					for (int m = 1; m < Q[q].size(); m++) {
						bool indBl = true;  // Indicates if microservice 'm' is parallelizable with its predecessor
						for (int mp : get<1>(Q[q][m])) {  // Iterate over the precedence list of microservice 'm'
							if (mp == get<0>(Q[q][m - 1])) {
								indBl = false;  // 'm' is not parallelizable if it includes microservice 'm-1'
							}
						}
						if (indBl) {
							marge++;  // Increase the margin if 'm' is parallelizable with its predecessor
						}
					}
					sf.SFC = Q[q];
					sf.score = marge;
					orderedSFC.push_back(sf);
				}
				return orderedSFC;
			}

			// Assigns a simple increasing index to each SFC based on the margin between required latency and execution latency
			vector<scoreSFC> attribuerIndiceApp(vector<vector<tuple<int, vector<int>>>> Q, vector<int> latenceDemande, vector<int> sources, vector<int> destinations) {
				vector<scoreSFC> orderedSFC;
				// Iterate through the list of SFCs
				for (int q = 0; q < Q.size(); q++) {
					scoreSFC sf;
					// Set the source, destination, and latency
					sf.id = q;
					sf.src = sources[q];
					sf.dst = destinations[q];
					sf.lat = latenceDemande[q];
					sf.SFC = Q[q];
					// Assign the index
					sf.score = q + 1;
					orderedSFC.push_back(sf);
				}
				return orderedSFC;
			}

			// Function to assign a fixed score to all SFCs
			vector<scoreSFC> attribuerScoreApp(vector<vector<tuple<int, vector<int>>>> Q, int scoreFixe, vector<int> latenceDemande, vector<int> sources, vector<int> destinations) {
				vector<scoreSFC> orderedSFC;
				// Iterate through the list of SFCs
				for (int q = 0; q < Q.size(); q++) {
					scoreSFC sf;
					// Set the source, destination, and latency
					sf.id = q;
					sf.src = sources[q];
					sf.dst = destinations[q];
					sf.lat = latenceDemande[q];
					sf.SFC = Q[q];
					// Assign the fixed score
					sf.score = scoreFixe;
					orderedSFC.push_back(sf);
				}
				return orderedSFC;
			}


			// Function to assign a specific score to each SFC based on scores provided as a parameter
			vector<scoreSFC> attribuerScorePermute(vector<vector<tuple<int, vector<int>>>> Q, vector<int> score, vector<int> latenceDemande, vector<int> sources, vector<int> destinations) {
				vector<scoreSFC> orderedSFC;
				// Iterate through the list of SFCs
				for (int q = 0; q < Q.size(); q++) {
					scoreSFC sf;
					// Assign the source, destination, and latency
					sf.src = sources[q];
					sf.dst = destinations[q];
					sf.lat = latenceDemande[q];
					sf.id = q;
					sf.SFC = Q[q];
					// Assign a score
					sf.score = score[q];
					orderedSFC.push_back(sf);
				}
				return orderedSFC;
			}

			// Function to deploy a microservice and return the deployment value (node, micro-service)
			bool deplSFcPathIndividual(vector<tuple<int, vector<int>>> path, vector<tuple<int, vector<int>>> SFC, vector<node>*nodeLis, int indM, vector<tuple<int, vector<int>, vector<vector<int>>>>*dep) {
				int iT = 0; // Index to traverse nodes
				bool depSI = true; // Indicates whether we have reached the end of the path (false) or not (true)

				// Calculate the margin of the path = Available space - size of the SFC
				int margeSFC;
				int memoryPath = 0;
				int lengthSFC = SFC.size();

				// Calculate the available capacity on a path
				for (int i = 0; i < path.size(); i++) {
					memoryPath += (*nodeLis)[get<0>(path[i])].capaMem;
				}

				margeSFC = memoryPath - lengthSFC;

				if (margeSFC >= 0) {
					if (get<0>(path[0]) != -1) {
						bool depI = false;
						while ((iT < path.size()) && (!depI)) {
							if ((*nodeLis)[get<0>(path[iT])].capaMem > 0) { // Deploy the microservice on node iT if possible
								bool ind = false;
								int idPlacementM = 0;
								// Check if node iT already exists in the deployment and get its ID
								while ((idPlacementM < (*dep).size()) && (!ind)) {
									if (get<0>((*dep)[idPlacementM]) == get<0>(path[iT])) {
										get<1>((*dep)[idPlacementM]).push_back(get<0>(SFC[indM]));
										ind = true;
									}
									if (!ind) {
										idPlacementM++;
									}
								}
								if (!ind) { // If node iT does not exist in the deployment, add it
									(*dep).push_back({ get<0>(path[iT]), {get<0>(SFC[indM])}, {} });
									idPlacementM = 0;
									while ((idPlacementM < (*dep).size()) && (!ind)) {
										if (get<0>((*dep)[idPlacementM]) == get<0>(path[iT])) {
											ind = true;
										}
										if (!ind) {
											idPlacementM++;
										}
									}
								}
								(*nodeLis)[get<0>(path[iT])].capaMem--; // Update the remaining capacity on node iT
								depI = true;
								// Additional logic for grouping parallelizable microservices (if necessary)
								// [ ... omitted for brevity ... ]
							}
							else {
								iT++;
							}
						}
						if (iT == path.size()) {
							depSI = false;
							return false;
						}
					}
					else {
						return false;
					}
				}
				else {
					return false;
				}
				return true;
			}


			// Enables parallel deployment of microservices for multiple SFCs (Service Function Chains)
			vector<deployment> deployementSPara(vector<scoreSFC> orderedSFC, vector<node>*listNode, int in) {
				vector<deployment> resDep(orderedSFC.size()); // Resulting deployment vector for each SFC
				bool notDep; // Flag to indicate if a microservice is not deployed
				bool solution; // Flag to indicate if there is a solution for deployment
				vector<int> lowPath; // Stores the shortest path for an SFC
				int sizeEppsteinTree = 0; // Size of the Eppstein tree for path calculations

				int src, dst; // Source and destination for the SFCs
				vector<tuple<int, vector<int>>> kPath = {}; // Temporary storage for a path
				vector<treeNode*> origTab(orderedSFC.size()); // Array to store the initial nodes for Eppstein algorithm for each SFC
				vector<MinHeap*> heapTab(orderedSFC.size()); // Array to store the heap for Eppstein algorithm for each SFC
				vector<vector<node>*> listNodeEPath(orderedSFC.size()); // List of nodes for each path

				// 1. Process each SFC:
				for (int q = 0; q < orderedSFC.size(); q++) {
					src = orderedSFC[q].src;
					dst = orderedSFC[q].dst;

					// Invert the graph for reverse path calculation
					listNodeEPath[q] = graphInverse2(*listNode);

					// Calculate the shortest path from DST to SRC using Dijkstra's algorithm
					djikstra(listNodeEPath[q], dst, src, 0);

					// Calculate the shortest path from SRC to DST and store in lowPath
					lowPath = lowPathF(*listNodeEPath[q], src, dst);

					// If a shortest path exists, set up Eppstein's algorithm structures
					if (!lowPath.empty() && lowPath.back() == dst) {
						listNodeEPath[q] = graphInverse2(*listNodeEPath[q]); // Invert the graph to correct direction
						eppsteinF(listNodeEPath[q]); // Run Eppstein algorithm for k shortest paths

						treeNode* orig = new treeNode(); // Tree node for Eppstein's algorithm
						sizeEppsteinTree = 0; // Reset size of the Eppstein tree
						*orig = eppsteinListNode(listNodeEPath[q], src, dst, &sizeEppsteinTree); // Initialize Eppstein tree

						heapTab[q] = new MinHeap(999); // Create a min heap for Eppstein's algorithm
						initMinHeap(heapTab[q], orig); // Initialize min heap with Eppstein's tree

						origTab[q] = orig; // Store the initial node for later use
					}
					else {
						// If no shortest path exists, mark the deployment as impossible for this SFC
						cout << "Deployment impossible (E2: No path between source and destination) for SFC: " << q << endl;
						resDep[q] = deployment{}; // Store an empty deployment
						heapTab[q] = nullptr; // Indicate that there's no heap for this SFC
					}
				}

				// Initialize arrays for k-shortest paths and deployment flags
				vector<vector<tuple<int, vector<int>>>> kPathTable(orderedSFC.size());
				vector<int> indMic(orderedSFC.size(), 0);
				vector<vector<tuple<int, vector<int>, vector<vector<int>>>>> depTab(orderedSFC.size());

				int t = 0; // Counter for processed SFCs
				bool indb = false; // Indicates if the current microservice has been deployed

				// Main loop for processing SFCs and deploying microservices
				while (t < orderedSFC.size()) {
					// Iterate through all SFCs for deployment
					for (int q = 0; q < orderedSFC.size(); q++) {
						// Check if there are still microservices to deploy for the current SFC and if it's not marked as completed
						if (indMic[q] < orderedSFC[q].SFC.size() && indMic[q] != -1) {
							// Attempt to deploy the current microservice
							indb = deplSFcPathIndividual(kPathTable[q], orderedSFC[q].SFC, listNode, indMic[q], &depTab[q]);
							if (indb) {
								// If deployment is successful, move to the next microservice
								indMic[q]++;
							}
							else {
								// If deployment fails, try the next path if one exists
								// First, remove any partial deployment and update node capacities
								for (int d = 0; d < depTab[q].size(); d++) {
									for (int md = 0; md < get<1>(depTab[q][d]).size(); md++) {
										(*listNode)[get<0>(depTab[q][d])].capaMem++; // Increment node capacity
									}
								}
								// Clear the current deployment attempt
								depTab[q].clear();

								// Attempt deployment on the next path
								bool indc = true; // Indicates if deployment was successful
								while (indc) {
									// Retrieve the next shortest path for the current SFC
									vector<tuple<int, vector<int>>> kPathOut = nextPath(origTab[q], heapTab[q], *listNodeEPath[q], kPathTable[q], orderedSFC[q].src, orderedSFC[q].dst);
									// If a next path exists, attempt deployment on it
									if (!kPathOut.empty()) {
										kPathTable[q] = kPathOut; // Store the new path
										indb = deployOnNewPath(kPathTable[q], orderedSFC[q].SFC, listNode, indMic[q], &depTab[q]);
										// If deployment is not successful, repeat the process with the next path
										if (!indb) {
											// Clean up any partial deployment
											clearDeployment(&depTab[q], listNode);
										}
										else {
											// Deployment is successful, break out of the loop
											indc = false;
										}
									}
									else {
										// No more paths are available for deployment
										indc = false;
										cout << "Deployment impossible (E1: No more paths available) for SFC: " << q << endl;
									}
								}
								// If we exit with indb == true, deployment is successful; otherwise, it's impossible
								if (!indb) {
									// Deployment impossible
									t++;
									indMic[q] = -1; // Mark the microservice as processed
									cout << "Error: Deployment impossible on alternate paths" << endl;
									// Clean up any partial deployment
									clearDeployment(&depTab[q], listNode);
								}
								else {
									// Successful deployment, prepare for the next microservice
									indMic[q]++;
								}
							}
						}
						// If all microservices for the current SFC are deployed, mark as complete
						if (indMic[q] == orderedSFC[q].SFC.size()) {
							t++; // Increment the counter for successful deployments
							indMic[q] = -1; // Mark the SFC as fully deployed
						}
					}
				}

				// Finalize the deployment for each SFC if possible
				for (int q = 0; q < orderedSFC.size(); q++) {
					if (!depTab[q].empty()) {
						// If deployment is possible, finalize the path and deployment details
						chaining(&kPathTable[q]);
						deployment d;
						d.chain = kPathTable[q];
						d.placement = depTab[q];
						resDep[q] = d;
					}
				}

				return resDep;
			}



